#!/usr/local/bin/perl
#
# PPSSUMM
#
# XMM Pipeline Processing System HTML summary generation script
#
# Author: Brendan Perry (bhp@star.le.ac.uk)
#
# Description: This script generates an HTML summary page(s) from
#              a PPS summary index fits file. 
# Usage: 
#  PPSSUMM infile="infilename" outfile="outfilename" logo="logo.gif" noclobber 
#
# noclobber and logofile switches are optional.
#

sub ppssumm {
    
    use English;
    use DAL;
    use SAS;
    use SAS::error;
    
    my ($taskname) = 'PPSSUMM';
    my ($date) = &date_time;
    my ($VERSION) = "2.2.7";
    my ($create) = "SAS task $taskname V $VERSION (Generated $date)";
    my ($msg,$code,$htmlfile);
    my ($header_ok) = 0;
    my ($ds,$fh,$top,$logo,$name,@commons,@allexts);
    my ($anchor,$epic_hash,$om_hash,$rgs_hash);
    
#
# parse_cl uses Perl Module GetOpts to parse the input command line
# infile=$infile outfile=$outfile instrument=$inst logo=$logo -noclobber
#
# $infile  = PPS summary FITS file
# $outfile = HTML output file
# $instrument = which instrument to do (PPS/EPIC/RGS/OM).
#
    my ($infile,$outfile,$inst,$logo,$noclobber) = &parse_cl();
    chomp ($inst,$infile);
    my $set_key = &get_key($infile);
    
    $in = $inst;
    %bandlabel=(0 => '--------',
		1 => '0.2-0.5',
		2 => '0.5-2.0',
		3 => '2.0-4.5',
		4 => '4.5-7.5',
		5 => '7.5-12.0',
		8 => '0.2-12.0');
    %short1 = ( "INSTRUMENT" => "Instrument",
		"EXP_DURATION" => "Duration",
		"EXP_START" => "Start time",
		"EXP_STOP" => "Stop time",
		"EXP_ID" => "ExpID",
		"DATAMODE" => "Mode", 
		"EPIC_FILTER" => "Filter",
		"OM_FILTER" => "Filter",
		"RGS_ORDER" => "Order"
		);
    
    %shorts = ( "HOUSEKEEPING" => "Housekeeping",
	        "NAME" => "Filename",
	        "CONTENT" => "Content",
	        "INSTRUMENT" => "Instrument",
	        "EPIC_FILTER" => "Filter",
	        "OM_FILTER" => "Filter",
	        "BAND" => "Band (keV)",
	        "OBSERVATION" => "Observation",
		"SENSITIVITY" => "SENS",
		"BOX-LOCAL" => "Box-Local",
		"SOURCE" => "Source",
		"SUMMARY" => "Summary",
		"IMAGE" => "Image",
		"EXPOSURE" => "Exposure",
		"FLUCTUATIONS" => "Fluctuations",
		"SPECTRUM" => "Spectrum",
		"BACKGROUND" => "Background",
		"TIMESERIES" => "Timeseries",
		"EVENT" => "Event",
		"INTENSITY" => "Intensity",
		"HISTOGRAM" => "Histogram",
		"IMAGING" => "Imaging",
		"TIMING" => "Timing",
		"BOX-MAP"=> "Box-Map",
		"FLATFIELD" => "Flatfield",
		"HISTORY" => "History",
		"PLOT" => "Plot",
		"TRAILED" => "Trailed ",
		"AUXILIARY" => "Auxiliary",
		"BRIGHT" => "Bright",
		"PRODUCT" => "Product",
		"MESSAGE" => "Message",
		"CATALOGUE" => "Catalogue",
		"CROSS-CORRELATION" => "Cross-Correlation",
		"CROSS-CORRELATIONS" => "Cross-Correlations",
		"CROSS-CORRELATI" => "Cross-Correlations",
		"DESCRIPTIONS" => "Descriptions",
		"FINDING" => "Finding",
		"POINTING" => "Pointing",
		"POINTED"=> "Pointed",
		"POSITION" => "Position",
		"SLEW" => "Slew",
		"TRACKING" => "Tracking",
		"PROCREVISION" => "Proc Revision",
		"DURATION" => "Duration",
		"OBJECT" => "Object",
		"SASVERSION" => "SAS Version",
		"PPSVERSION" => "PPS Version",
		"OBS_MODE" => "Mode",
		"DATAMODE" => "Inst&nbsp; Mode",
		"OBS_START" => "Start time",
		"OBS_STOP" => "Stop time",
		"OBS_DURATION" => "Duration",
		"EXP_DURATION" => "Duration",
		"EXP_START" =>    "Start time",
		"EXP_STOP" =>     "Stop time",
		"EXP_ID" => "Exposure&nbsp;ID",
		"OBSERVER" => "Observer",
		"VVFLAGS" => "V&amp;V Flags",	     	    
		"RGSORDER" => "Order",
		"RA_PNT" => "RA",
		"DEC_PNT" => "Dec",
	        "OCB" => "OCB"
		);

# Regexp for matching instrument to target
    $instmatch={'EPIC' => 'EMOS1|EMOS2|EPN',
		'RGS' => 'RGS1|RGS2',
		'OM' => 'OM'};
#
# epic columns: 'top' are commons for section heads
#
    (@EPICOBS) = ('NAME','CONTENT','OBS_MODE','RA_PNT','DEC_PNT','OBS_START','OBS_STOP','OBS_DURATION');   
    (@eotop) = (5,6,7,3,4,2);
    (@eobot) = (0,1);
    
    (@EPICEXP) = ('NAME','CONTENT','INSTRUMENT','RA_PNT','DEC_PNT','DATAMODE','EPIC_FILTER','EXP_ID','EXP_START','EXP_STOP','EXP_DURATION','BAND','VVFLAGS');
    (@eetop) = (2,7,5,6,8,10,9);
    (@eebot) = (0,1,11,12);
#    
# OM columns:
#
    (@OMOBS) = ('NAME','CONTENT','OBS_MODE','RA_PNT','DEC_PNT','OBS_START','OBS_STOP','OBS_DURATION');   
    (@ootops) = (0,1,2,3,4,5,6,7);
    (@oobot) = (0,1);

    (@OMEXP) = ('NAME','CONTENT','INSTRUMENT','RA_PNT','DEC_PNT','DATAMODE','OM_FILTER','EXP_ID','EXP_START','EXP_STOP','EXP_DURATION','VVFLAGS');    
    (@oetop) = (2,7,5,6,8,10,9);
    (@oebot) = (0,1,11);

    (@OMSW) = @OMEXP;

#
# rgs columns:
#
    (@RGSOBS) = ('NAME','CONTENT','OBS_MODE','RA_PNT','DEC_PNT','OBS_START','OBS_STOP','OBS_DURATION');   
    (@rotop) = (5,6,7,3,4);
    (@robot) = (0,1);

    (@RGSEXP) = ('NAME','CONTENT','INSTRUMENT','RA_PNT','DEC_PNT','DATAMODE','OCB','EXP_ID','EXP_START','EXP_STOP','EXP_DURATION','VVFLAGS');    
    (@retop) = (2,7,5,8,10,9);
    (@rebot) = (0,1,11);

    (@RGSSRC) = @RGSEXP;
#
# cat columns
#
# catobs = @commons
#
    (@CATOBS) = ('NAME','CONTENT','OBS_MODE','RA_PNT','DEC_PNT','OBS_START','OBS_STOP','OBS_DURATION');   
    (@cotop) = (5,6,7,3,4);
    (@cobot) = (0,1);

    (@CATSRC) = ('NAME','CONTENT','OBS_MODE','RA_PNT','DEC_PNT','PLATE_RA','PLATE_DEC','PLATE_FILTER','PLATE_ID','PLATE_ORIGIN','SRC_NUM');
    (@cetop) = (2,7,5,6,8,9,10);
    (@cebot) = (0,1,10);
#
# ppsobs
# 
    (@PPSOBS) = ('NAME','CONTENT');
    (@PPSEXP) = ('INSTRUMENT','EXP_ID','DATAMODE','EXP_START','EXP_DURATION','EXP_STOP');

    (@INSTSUMM) = ('INSTRUMENT','ACTIVE','NEXPOSURES','PRIORITY');

    my (%insts) = ('EMOS1','EPIC','EMOS2','EPIC','EPN','EPIC','RGS1','RGS','RGS2','RGS','OM','OM');
    my ($yes) = 1;            # append to existing file in open_file
    my ($no,$problem) = 0;    # overwrite existing file in open_file
    my ($tlog,$message);
    my (@outlog,@badlog);
    my ($date) = &date_time;
#
# open the PPS summary fits file using peDAL
#
    chomp $infile;
    my ($ds) = &open_fits($infile,"READONLY");
#
# Read the primary header
#
#
# Read Attributes
# At time of writing the TARGET keyword is absent from the input files.  It included for completeness.
#
    my (%attributes) = (
			 'OBS_ID'   => { value => "" , text => "Observation"} 
			, 'ORBIT'    => { value => "" , text => "Revolution"} 
			, 'DATE-OBS' => { value => "" , text => "Start Time"} 
			, 'DATE-END' => { value => "" , text => "End Time"} 
			, 'OBSERVER' => { value => "" , text => "Observer"}
			, 'RA_NOM'   => { value => "" , text => "RA"} 
			, 'DEC_NOM'  => { value => "" , text => "Dec"}
			, 'OBJECT'   => { value => "" , text => "Target"}
			);
    my (@attr_order) = ("ORBIT" , "OBJECT" ,"OBSERVER");
    
    &read_attributes($ds,\%attributes);
    my $sub_title = &generate_sub_title(\%attributes,\@attr_order);
    my $title="$attributes{OBS_ID}{value} $inst Processing Summary";
#
# read the PPS summary fits file data using peDAL
#
    my ($file,$val,$num);
    ($extension,$var,%hash_ref);
    
    if ($inst =~ /EPIC/) {
	(@allexts) = ("EPICEXP");
    } elsif ($inst =~ /OM/) {
	(@allexts) = ("OMEXP","OMSW");
    } elsif ($inst =~ /RGS/) {
	(@allexts) = ("RGSEXP","RGSSRC");
    } elsif ($inst =~ /PPS/) {
	(@allexts) = ("EPICEXP","OMEXP","OMSW","RGSEXP","RGSSRC");
    } elsif ($inst =~ /CAT/) {
	(@allexts) = ("CATSRC");
    } else {
	$msg = "Instrument page selected not one of EPIC|RGS|OM|CAT|PPS";
	$code = "InvalidInstEntered";
	chomp ($msg,$code);
      SAS::error($code,$msg);    # shouldn't really ever get to here
    }
#
# read_extension is done for EACH value of @allexts above (or rather,
# the ARRAY of that name, e.g. @EPICOBS)
#
# call read_extension($ds,$extension=="EPICOBS",@$extension==@EPICOBS)
#
# get back @ext_ref array, @ext_ref[nfiles][ncolumns], with
# filename in $ext_ref[row][0]
#    
# rename through symbolic ref -> %epicobs = $hash_ref
# creates a new hash with the lowercase name of the array.
#
# e.g. @EPICOBS is sent to read_extension, which reads the 
#      columns contained in @EPICOBS from the extension EPICOBS of
#      the ppssummary index file.
#      The array returned is then renamed to @epicobs
    
    foreach $extension (@allexts) {  # only extensions that need to be read
      SAS::message($SAS::AppMsg,$SAS::SparseMsg,"Reading $extension extension in $infile.");
	($ext_ref,$names) = &read_extension($ds,$extension,@$extension);
	$var = lc($extension);    # e.g. $var becomes $epicobs
	@{$var} = @$ext_ref;      # e.g. @epicobs = $ext_ref
    }
    if (defined(@omsw)) {
	push(@omexp,@omsw);
    }
    if (defined(@rgssrc)) {
	push(@rgsexp,@rgssrc);
    }
    undef(@omsw);
    undef(@rgssrc);
#
# now get ALL instrument's OBS extension (to make links on each page)
# the filename, etc for each instrument summary file is contained
# in the OBS extension for that instrument.
#
    my (@obsexts) = ("EPICOBS","OMOBS","RGSOBS","PPSOBS","CATOBS");
    
    foreach $extension (@obsexts) {
        SAS::message($SAS::AppMsg,$SAS::SparseMsg,"Reading $extension extension in $infile.");
	    ($ext_ref,$names) = &read_extension($ds,$extension,@$extension);
	    push (@summnames,@$names);  # summary filenames
		$var = lc($extension);    # e.g. $var becomes $epicobs
	    @{$var} = @$ext_ref;      # e.g. @epicobs = $ext_ref
    }
#
# check the GIF logo file for existence and readability
#
    my ($logo_ok) = &check_logo($logo);
    if ($logo_ok) {
      SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Checked GIF logo: $logo ...ok");
    } else {
	$msg = "No LOGO to be written to HTML file.";
	$code = "NoLogoBanner";
	chomp ($msg,$code);
      SAS::warning($code,$msg);    # warn
    }
#
# If you at least have an instrument summary, can write a
# (bad looking) html file:
#
# open HTML file and write header
#
    ($htmlfile) = &open_file($outfile,"WRITEONLY");
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing HTML Header.");
    ($header_ok) = &write_HTML_header($htmlfile,$title,$sub_title,$logo,$logo_ok,@summnames);
    
#
# write instsumm records with links to other pages
#
#    if (@instsumm) {
#      SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing Instrument Summary section.");
#	&write_HTML_instsumm($htmlfile,\@instsumm,\@INSTSUMM,$inst);
#	&glossary_($htmlfile);
#    } else {
#	$msg = "No INSTSUMM records written to HTML.";
#	$code = "NoINSTSUMMtoHTML";
#	chomp ($msg,$code);
#      SAS::warning($code,$msg);    # warn
#    }
#
# write ppsobs records
#
    if ($inst =~ /PPS/) {
	(@ppbot) = ("NAME","CONTENT");
#      SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Entering read_exp @allexts");
	($pps_ref) = &read_exp($ds,\@allexts,\@PPSEXP);
      SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing PPS Summary section.");
	&write_HTML_pps($htmlfile,$pps_ref,$set_key,\@PPSOBS,\@ppsobs,\@summnames,\@PPSEXP);
    } else {
	$msg = "No PPSOBS records written to HTML.";
	$code = "NoPPSOBStoHTML";
	chomp ($msg,$code);
      SAS::warning($code,$msg);    # warn
    }
 
# close the fits file now that values are in hashes
#
  SAS::message($SAS::AppMsg,$SAS::SparseMsg,"Releasing Index File dataset.");
    $ds->release;
#
# write EPIC records
#
#
    if ($inst =~ /EPIC/) {
#
# write epicobs records
#
	if (@epicobs) {
	  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing EPICOBS Summary section.");
	    &write_HTML_multi($htmlfile,\@epicobs,\@EPICOBS,\@eotop,\@eobot,0,"EPIC Multi-Exposure Products");
	} else {
	    $msg = "No EPICOBS records written to HTML.";
	    $code = "NoEPICOBStoHTML";
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);    # warn
	}
#
# write epicexp records
#
	if (@epicexp) {
	  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing EPICEXP Summary section.");
	    &write_HTML_table($htmlfile,\@epicexp,\@EPICEXP,\@eetop,\@eebot,2,7,0,"EPIC Exposures Processed by PPS",7,2);
	} else {
	    $msg = "No EPICEXP records written to HTML.";
	    $code = "NoEPICEXPtoHTML";
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);    # warn
	}
#
# write epicsrc records
#
#	if (@epicsrc) {
#	  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing EPICSRC Summary section.");
#	    &write_HTML_table($htmlfile,\@epicsrc,\@EPICSRC,\@estop,\@esbot,0,3,4,"EPIC Source Summary");
#	} else {
#	    $msg = "No EPICSRC records written to HTML.";
#	    $code = "NoEPICSRCtoHTML";
#	    chomp ($msg,$code);
#	  SAS::warning($code,$msg);    # warn
#	}
    }
#
# write OM records
#
    if ($inst =~ /OM/) {
#
# write omobs records
#
	if (@omobs) {
	  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing OMOBS Summary section.");
	    &write_HTML_multi($htmlfile,\@omobs,\@OMOBS,\@ootop,\@oobot,0,"OM Multi-Exposure Products");
	} else {
	    $msg = "No OMOBS records written to HTML.";
	    $code = "NoOMOBStoHTML";
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);    # warn
	}
#
# write omexp records
#
	if (@omexp) {
	  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing OMEXP Summary section.");
	    &write_HTML_table($htmlfile,\@omexp,\@OMEXP,\@oetop,\@oebot,2,7,0,"OM Exposures Processed by PPS",7,2);
	} else {
	    $msg = "No OMEXP records written to HTML.";
	    $code = "NoOMEXPtoHTML";
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);    # warn
	}
    }
#
# write omsrc records
#
#	if (@omsrc) {
#	  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing OMSRC Summary section.");
#	    &write_HTML_table($htmlfile,\@omsrc,\@OMSRC,\@ostop,\@osbot,0,8,9,"OM Source Summary");
#	} else {
#	    $msg = "No OMSRC records written to HTML.";
#	    $code = "NoOMSRCtoHTML";
#	    chomp ($msg,$code);
#	  SAS::warning($code,$msg);    # warn
#	}
    
#
# write RGS records
#
    if ($inst =~ /RGS/) {
#
# write rgsobs records
#
	if (@rgsobs) {
	  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing RGSOBS Summary section.");
	    &write_HTML_multi($htmlfile,\@rgsobs,\@RGSOBS,\@rotop,\@robot,0,"RGS Multi-Exposure Products");
	} else {
	    $msg = "No RGSOBS records written to HTML.";
	    $code = "NoRGSOBStoHTML";
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);    # warn
	}    
	#
	# write rgsexp records
	#
	if (@rgsexp) {
	  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing RGSEXP Summary section.");
	    &write_HTML_table($htmlfile,\@rgsexp,\@RGSEXP,\@retop,\@rebot,2,7,0,"RGS Exposures Processed by PPS",7,2);
	} else {
	    $msg = "No RGSEXP records written to HTML.";
	    $code = "NoRGSEXPtoHTML";
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);    # warn
	}
	#
	# write rgssrc records
	#
#	if (@rgssrc) {
#	  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing RGSSRC Summary section.");
#	    &write_HTML_table($htmlfile,\@rgssrc,\@RGSSRC,\@rstop,\@rsbot,0,7,8,"RGS Source Summary");
#	} else {
#	    $msg = "No RGSSRC records written to HTML.";
#	    $code = "NoRGSSRCtoHTML";
#	    chomp ($msg,$code);
#	  SAS::warning($code,$msg);    # warn
#	}
    }
#
# write CAT records
#
    if ($inst =~ /CAT/) {
#
# write catobs records
#
	if (@catobs) {
	  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing CATOBS Summary section.");
	    &write_HTML_multi($htmlfile,\@catobs,\@CATOBS,\@cotop,\@cobot,0,"Catalogue Multi-Exposure Products");
	} else {
	    $msg = "No CATOBS records written to HTML.";
	    $code = "NoCATOBStoHTML";
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);    # warn
	}
	#
	# write catcat records
	#
	if (@catcat) {
	  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing CATCAT Summary section.");
	    &write_HTML_table($htmlfile,\@catcat,\@CATCAT,\@cctop,\@ccbot,0,0,0,"CAT Catalog Summary");
	} else {
	    $msg = "No CATCAT records written to HTML.";
	    $code = "NoCATCATtoHTML";
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);    # warn
	}
	#
	# write catsrc records
	#
	if (@catsrc) {
	  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing CATSRC Summary section.");
	    &write_HTML_table($htmlfile,\@catsrc,\@CATSRC,\@cstop,\@csbot,0,0,10,"CAT Source Summary");
#	&top_($htmlfile);
#	&glossary_($htmlfile);
	} else {
	    $msg = "No CATSRC records written to HTML.";
	    $code = "NoCATSRCtoHTML";
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);    # warn
	}
    }
#
# add the glossary of abbreviations
#
#    &write_glossary($htmlfile);
&top_($htmlfile);
#
# create and write footer to HTML page
#
&write_HTML_footer($htmlfile,$create);
 SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Wrote HTML footer.");
#
# close the open and written HTML file:
#
$htmlfile->close;
 SAS::message($SAS::AppMsg,$SAS::SparseMsg,"Closed HTML filehandle.");    
1;
}
#
#=========================================================================
#
# PPSSUMM subroutine section
#
#=========================================================================
# --
# PARSE_CL
# --
sub parse_cl {       # parse the command line with pedal
# infile = Input FITS PPS index filename
# outfile = HTML output filename
# instrument = inst name (EPIC,RGS,OM)
# logofile = GIF logo 
#
# Command line has the form:
#
# ppssumm infile="ppssummary.FITS" outfile="file.HTM" inst="instrument"
#         -noclobber [-v -h]
#
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Parsing command line.");
    
    my ($in) = stringParameter("set");
    my ($out) = stringParameter("outfile");
    my ($inst) = stringParameter("page");
    my ($logo) = stringParameter("logofile");
    my ($clob)= DAL::clobber();
    my ($msg,$code);
    
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"noclobber is $clob.");
    
    chomp($inst);
    $inst = uc($inst);
    
    if (!($inst =~ /(EPIC)|(RGS)|(OM)|(CAT)|(PPS)/)) {
	$msg = "The page parameter must be EPIC|RGS|OM|CAT|PPS";
	$code = "InvalidPageEntered";
	chomp ($msg,$code);
      SAS::error($code,$msg);    # exit with error
    }
    
    chomp ($in,$out,$inst,$logo);
    return ($in,$out,$inst,$logo,$clob);
}
# --
# OPEN_FILE
# --
sub open_file {   # open an HTML output file

    require IO::Handle;
    require FileHandle;
    
    my ($file,$mode) = @_;
    my ($msg) = "Unable to open file: $file.";
    my ($code);
    my ($clob) = DAL::clobber();    
    my ($fh) = new FileHandle;
    
    if ($fh) {
	if ($mode =~ /READ/) {              # open for read ONLY
	    $file = "<$file";  
	    if (!($fh->open("$file"))) {
		$code = "RdOnlyOpenFailure";
		chomp ($msg,$code);
	      SAS::error($code,$msg);    # exit with error
	    } else {
	      SAS::message($SAS::AppMsg,$SAS::NoisyMsg,"Opened file $file READONLY");
	    }
	} elsif ($mode =~ /WRITE/) {        # open for write ONLY
	    if (-e $file) {
	      SAS::message($SAS::AppMsg,$SAS::NoisyMsg,"Output file already exists.");
		if ( $clob ) {    # noclobber set to TRUE
		    $msg = "File exists but noclobber NOT set: overwriting.";
		    $code = "ClobberingFile";
		    chomp ($msg,$code);
		  SAS::warning($code,$msg);    # warn

		} 
		else {
		    # noclobber set to TRUE
		    $msg = "Cannot open WRITE with noclobber set.";
		    $code = "NoclobberOpenFailure";
		    chomp ($msg,$code);
		  SAS::error($code,$msg);    # exit with error
		}
	    }
	    $file = ">$file";  # overwrite
	    if (!($fh->open("$file"))) {
		$msg = "Failure to open file writeonly";
		$code = "WrtOnlyOpenFailure";
		chomp ($msg,$code);
	      SAS::error($code,$msg);    # exit with error
	    } else {
	      SAS::message($SAS::AppMsg,$SAS::NoisyMsg,"Opened file $file WRITEONLY");
	    }
	}
	return ($fh);
    } else {
	$msg = "Unable to get new FileHandle";
	$code = "FileHandleFailure";
	chomp ($msg,$code);
      SAS::error($code,$msg);    # exit with error
    }
}
#
# end OPEN_FILE
#
# --
# CHECK_LOGO
# --
#
sub check_logo {   # check input logo file 

    my ($logo) = @_;
    my ($msg,$code,$ok);
#
# check the input logo file for existence, readability and
# if it's a Binary file. Not very vigorous test, but ok for now
# (if you're DUMB enough to put a non-gif/tiff/png file as
# a logo banner you get what you get!).
#
    if ($logo) {
	if (!(-e $logo)) {
	    $msg = "File: $logo does not exist.";
	    $code = "LogoFileNotExist";
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);			
	    $ok = 0;
	} elsif (!(-r $logo)) {
	    $msg = "File: $logo not readable.";
	    $code = "NoReadLogoFile";
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);			
	    $ok = 0;
	} elsif (!(-B $logo)) {
	    $msg = "File: $logo possibly not binary image file.";
	    $code = "LogoFileNotImage";
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);			
	    $ok = 0;
	} else {
	    $ok = 1;    # file should be ok
	}
    } else {
	$ok = 0;
    }
    return ($ok);
}
#
# end check_logo
#
# --
# WRITE_HTML_HEADER
# --
sub write_HTML_header {       # start HTML file with HEAD,TABLE,BODY,etc

#
# subroutine to write the header information to the output www page.
# this includes the TITLE, colors, top anchor, etc
#
    my ($fh,$title,$sub_title,$logo,$logo_ok,@summnames) = @_;
    chomp ($title,$logo,$logo_ok);
    
    my ($top) = "top";    # set anchor at top, links with bottom
    my ($bg) = "bgcolor=\"\#FFFFFF\"";   # white
    my ($text) = "text=\"\#0000EE\"";    # black
    my ($link) = "link=\"\#0000EE\"";    # black
    my ($vlink) = "vlink=\"\#551A8B\"";  # blueish?
    my ($alink) = "alink=\"\#FF0000\"";  # red?

    &write_generic($fh,"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"); 
    &write_generic($fh,"<!DOCTYPE html\n");
    &write_generic($fh,"PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"\n");
    &write_generic($fh,"\"http://www.w3.org/TR/xhtml1/DTD/xhtml-transitional.dtd\">\n");

    &write_generic($fh,"<html>\n");
    &write_generic($fh,"<head><title>$title</title></head>\n");
    &write_generic($fh,"<body $bg $link $vlink $alink>\n");
    if ($logo_ok) {
	my ($imgsrc) = "<img src=\"$logo\" alt=\"XMM Survey Science Centre\" />";
	&write_generic($fh,"$imgsrc\n");
    }
    &write_generic($fh,"<hr noshade=\"noshade\" />\n");

#
# write links to other summary file pages (if defined)
#
    if (@summnames) {
        SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Writing Links to other SUMMARY pages.");
	    &write_HTML_links($fh,$inst,@summnames);
    } else {
	    $msg = "No Instrument Summary Files found in OBS extensions.";
	    $code = "NoLinkstoSUMMFiles";
	    chomp ($msg,$code);
        SAS::warning($code,$msg);    # warn
    }

    &write_generic($fh,"<hr noshade=\"noshade\" />\n");
    my ($temp) = &anchor($top);
    &write_generic($fh,"$temp<h1>$title</h1>\n");
    &write_generic($fh,$sub_title);
    &write_generic($fh,"<br />\n");
    &write_generic($fh,"<hr noshade=\"noshade\" />\n");

}
#
# end WRITE_HTML_HEADER
#
# --
# WRITE_HTML_FOOTER
# --
sub write_HTML_footer {  # write the HTML footer info to the page bottom

    my ($fh,$create) = @_;
    my ($top) = "#rTOP";
    &write_generic($fh,"<br />\n");
    &write_generic($fh,"<h5>$create</h5>\n");
    &write_generic($fh,"</body>\n");
    &write_generic($fh,"</html>\n");
}
#
# end WRITE_HTML_FOOTER
#
# --
# top_
# --
sub top_ {

    my ($fh) = @_;
    my ($top) = "#rTOP";
    my ($line) = href($top,'Top of Page');  # reference to top anchor
    &write_generic($fh,"$line\n");	
    &write_generic($fh,"<br />\n");       
}
#
# end top_
#
# --
# glossary_
# --
sub glossary_ {

    my ($fh) = @_;
    my ($top) = "#GLOSSARY";
    my ($line) = href($top,'Glossary of abbreviations'); # ref to glossary
    &write_generic($fh,"$line\n");	
    &write_generic($fh,"<br />\n");       
}
#
# end glossary_
#
# --
# write_glossary
# --
sub write_glossary {

  my ($fh) = @_;
  my ($anch) = "GLOSSARY"; # set anchor at bottom, links with glossary
  my ($temp) = &anchor($anch);
  $temp .= "Glossary of abbreviations";
  chomp ($temp);
  
  my (%shorts) = ("AUX" => "AUXILIARY",
		  "B-L" => "BOX-LOCAL",
		  "B-M" => "BOX-MAP",
		  "BKG" => "BACKGROUND",
		  "BR" => "BRIGHT",
		  "CAT" => "CATALOGUE",
		  "D_MODE" => "DATAMODE",
		  "DESCR" => "DESCRIPTIONS",
		  "E_DUR" => "EXP_DURATION",
		  "EVT" => "EVENT",
		  "EXP" => "EXPOSURE",
		  "FIND" => "FINDING",
		  "FLTFLD" => "FLATFIELD",
		  "FLUCT" => "FLUCTUATIONS",
		  "HISTO" => "HISTOGRAM",
		  "HK" => "HOUSEKEEPING",
		  "HSTY" => "HISTORY",
		  "IMG" => "IMAGING or IMAGE",
		  "INTENS" => "INTENSITY",
		  "ML" => "MAX LIKELIHOOD",
		  "MSG" => "MESSAGE",
		  "O_DUR" => "OBS_DURATION",
		  "O_MODE" => "OBS_MODE",
		  "OBJ" => "OBJECT",
		  "OBS" => "OBSERVATION",
		  "OSW" => "OM SCIENCE WINDOW",
		  "P" => "POINTING or POINTED",
		  "PI" => "OBSERVER",
		  "PLT" => "PLOT",
		  "POS" => "POSITION",
		  "PPS" => "PPSVERSION",
		  "PROD" => "PRODUCT",
		  "REV" => "PROCREVISION",
		  "S" => "SLEW",
		  "SAS" => "SASVERSION",
		  "SENS" => "SENSITIVITY",
		  "SPEC" => "SPECTRUM",
		  "SRC" => "SOURCE",
		  "SUMM" => "SUMMARY",
		  "TIMING" => "TIMING",
		  "TRK" => "TRACKING",
		  "TRLD" => "TRAILED",
		  "TS" => "TIMESERIES",
		  "X-C" => "CROSS-CORRELATION",
		 );
  
  &begin_table($fh,$temp);  
  &write_generic($fh,"<TR>\n");
  &table_row($fh,"Abbrev","B","CENTER");
  &table_row($fh,"Definition","B","CENTER");
  &table_row($fh,"Abbrev","B","CENTER");
  &table_row($fh,"Definition","B","CENTER");
  &table_row($fh,"Abbrev","B","CENTER");
  &table_row($fh,"Definition","B","CENTER");
  &write_generic($fh,"</tr>\n");
  my ($i) = 0;
  &write_generic($fh,"<tr>\n");
  
  foreach $key (sort keys %shorts) {
    $i++;
    chomp $key;
    &table_row($fh,$key,"I","CENTER");	
    $name = $shorts{$key};
    chomp $name;
    &table_row($fh,$name,"B","CENTER");	
    if ($i == 3) {
      &write_generic($fh,"</tr>\n");
      $i = 0;
      &write_generic($fh,"<tr>\n");
    }
  }
  
  my ($line) = "</table>\n";
  &write_generic($fh,$line);
  
}
#
# end write_glossary
#
sub anchor {           # create an anchor from input
  my ($name) = @_;
  my ($anchor) = "<a name=\"$name\"></a>";
  chomp $anchor;
  return ($anchor);
}

sub href {             # create an href from input

# <a href="obssumm.html">Observation Summary Page</a>

  my ($ref,$name) = @_;
  my ($href) = "<a href=\"$ref\">$name</a>";
#  chomp $href;
  return ($href);
}

sub begin_table {      # start a new table

# <h2>Table Name:</h2> 

  my ($fh,$name) = @_;
  chomp ($name);
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Begin Table: $name");
  &write_generic($fh,"<h2>$name</h2>");
  my ($line) = "<table border>\n";
  &write_generic($fh,$line);
}

sub end_table {        # complete a table

  my ($fh,$name) = @_;
  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"End Table: $name");
  my ($line) = "</table>\n";
  &write_generic($fh,$line);
}

sub table_row {        # write a single table row

  my ($fh,$value,$type,$align) = @_;   # value, typeface B align CENTER
  chomp ($value,$type,$align);
  my ($tr) = "<td align=\"$align\"><$type>$value</$type></td>\n";
  &write_generic($fh,$tr);
}

sub date_time {          # create the file creation datestring

    my $time = localtime(time);
    my @time = split /\ /, $time;
    return "$time[2] $time[1] $time[4]";
}

#--
# WRITE_GENERIC
#--
sub write_generic {        # write a created HTML line

  my ($fh,$line) = @_;
  my ($msg) = "Failure to write to output HTML file";
  my ($code) = "WriteHTMLFailure";
  chomp ($msg,$code);
#
# exit with error if problem with printf
#
  if (!( printf ($fh $line) ) ) {
    SAS::error($code,$msg);
  }
}
#--
# open_fits
#-- 
sub open_fits {            # open FITS file $in_file for reading using peDAL

  my ($infile) = @_;
  my ($code,$msg);
  chomp $infile;
  
  my $ds = DAL::dataset->new(name => $infile, mode => $DAL::READ);
  if (defined ($ds)) {
    SAS::message($SAS::AppMsg,$SAS::SparseMsg,"Opened dataset: $infile");
  } else { 
    $code = "DatasetOpenFail";
    $msg = "Dataset Open Failure!";
    chomp ($code,$msg);
    SAS::error($code,$msg);
  }
  return($ds);
}
#
# end open_fits
#
#--
# READ_EXTENSION
#--
sub read_extension { # read a single FITS file extension into a hash

    use DAL::utils qw/setStringCell getStringCell/;
    
    my ($ds,$extname,@cols) = @_;
    my ($col, $v, $co , $v2 , $co2, $rows, $i, $elems, $type);
    my ($tb, @instlist, @new);
    my ($code,$msg,$name,$key,$temp);
    my $ncols = scalar(@cols);

#  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Checking for table $extname");

    if ($ds->hasTable(name => $extname) ) {   # does table exists?
#      SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Reading Table $extname");
	$tb = $ds->table(name => $extname);   # read correct table
	#
	# loop through and get all the columns 
	#
	$rows = $tb->rows();

	for ( $i=0 ; $i < $rows ; $i++ ) {		
	    for ($j=0;$j<$ncols;$j++) {
		$col = $cols[$j];
#	      SAS::message($SAS::AppMsg,$SAS::SparseMsg,"col is $col");
		if ($tb->hasColumn(name => $col)) {
		    $co2 = $tb->column(name => $col);
		    $v2 = $co2->data();
		    $type = $co2->type();
		    $ielems = $co2->elements();	    
		    #
		    # types: 1=bool, 2=Int8, 3=Uint16, 4=Int16,
		    #        5=Uint32, 6=Int32, 7=Real32, 8=Real64, 
		    #        9=DString
		    #
		    # 	if ($type == 7) { # real32
		    #	    $elems = $co2->elements();
		    #	    $new{$name}{$col} = $v2->value($i*$elems);
		    #
		    if ($type == 9) {  # string column
			$new[$i][$j] = getStringCell($co2,$i);
			if ( $col =~ /NAME/) {
			    if ($new[$i][$j] =~ /SUMMAR|PPSSUM|PPSMSG|CORRE/) {
			      SAS::message($SAS::AppMsg,$SAS::SparseMsg,"Summary file is $new[$i][$j]");
				push (@names , $new[$i][$j]);
			    }
			    if ($new[$i][$j] =~ /SUMMAR/) {
				$summs{$extname} = $new[$i][$j];
			    }
			}
		    } elsif (($type == 7) or ($type == 8)) {
			$temp = sprintf("%g",$v2->value($i));
#	      $new{$name}{$col} = $temp;
			$new[$i][$j] = $temp;
		    } else {
#	      $new{$name}{$col} = $v2->value($i);
			$new[$i][$j] = $v2->value($i);
		    }
		    $co2->release();   # release current column
		} else {
		    $msg = "No ".$col." column in ".$extname." extension, table will NOT be written";
		    $code = "No".$col."Column";
		    chomp ($msg,$code);
		  SAS::warning($code,$msg);
		}
	    }
	}
	$tb->release();
    } else {
	$msg = "No ".$extname." extension in file, table will NOT be written";
	$code = "No".$extname;
	chomp ($msg,$code);
      SAS::warning($code,$msg);
    }
    return (\@new,\@names);
}
#
# end read_extension
#
#--
# READ_EXP
#--
sub read_exp { # read exposure extensions, return array to print for ppsobs

    use DAL::utils qw/setStringCell getStringCell/;    
    my ($ds,$exts,$cols) = @_;   # exts = all exp exts, cols = cols to print
    my ($col, $v, $co , $c , $v2 , $co2, $rows, $i, $elems, $type);
    my ($tb, @instlist, @new);
    my ($code,$msg,$name,$key,$temp);
    my $ncols = scalar(@cols);
    my $offset=0;
    foreach $extname (@$exts) {
#      SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Extracting from $extname");
	if ($ds->hasTable(name => $extname) ) {   # does table exists?
#	  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Reading Table $extname");
	    $tb = $ds->table(name => $extname);   # read correct table
	    #
	    # loop through and get all the columns 
	    #
	    $c = 0;
	    foreach $col (@$cols) {
#	      SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Reading Column $col");
		$rows = $tb->rows();	    
		if ($tb->hasColumn(name => $col)) {
		    $co2 = $tb->column(name => $col);
		    $v2 = $co2->data();
		    $type = $co2->type();
		    $ielems = $co2->elements();	    
		    for ( $i=0 ; $i < $rows ; $i++ ) {		
			#
			# types: 1=bool, 2=Int8, 3=Uint16, 4=Int16,
			#        5=Uint32, 6=Int32, 7=Real32, 8=Real64, 
			#        9=DString
			#
			# 	if ($type == 7) { # real32
			#	    $elems = $co2->elements();
			#	    $new{$name}{$col} = $v2->value($i*$elems);
			#
			if ($type == 9) {  # string column
			    $new[$i+$offset][$c] = getStringCell($co2,$i);
			} elsif (($type == 7) or ($type == 8)) {
			    $temp = sprintf("%g",$v2->value($i));
			    $new[$i+$offset][$c] = $temp;
			} else {
			    $new[$i+$offset][$c] = $v2->value($i);
			}
			$co2->release();   # release current column

		    }

		} else {
		    $msg = "No ".$col." column in ".$extname." extension";
		    $code = "No".$col."Column";
		    chomp ($msg,$code);
		  SAS::warning($code,$msg);
		    for ( $i=0 ; $i < $rows ; $i++ ) {		
			$new[$i][$c] = "---";  
		    }
		}
		$c++;
	    }

	    $tb->release();
	} else {
	    $msg = "No ".$extname." extension in file, no PPS links written";
	    $code = "No".$extname;
	    chomp ($msg,$code);
	  SAS::warning($code,$msg);
	}
	if ($ds->hasTable(name => $extname) ) {$offset+=$rows;}
    }

    return (\@new);
}
#
# end read_pps
#
#--
# READ_BOT
#--
sub read_bot { # read exposure extensions, return array to print for ppsobs

    use DAL::utils qw/setStringCell getStringCell/;    
    my ($ds,$cols) = @_;   # exts = all exp exts, cols = cols to print
    my ($col, $v, $co , $v2 , $co2, $rows, $i, $elems, $type);
    my ($tb, @instlist, @new);
    my ($code,$msg,$name,$key,$temp);
    my $ncols = scalar(@cols);
    my $extname = "PPSOBS";

  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Extracting from $extname");

    if ($ds->hasTable(name => $extname) ) {   # does table exists?

      SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Reading Table $extname");

	$tb = $ds->table(name => $extname);   # read correct table
	#
	# loop through and get all the columns 
	#
	$c = 0;
	foreach $col (@$cols) {
	  SAS::message($SAS::AppMsg,$SAS::VerboseMsg,"Reading Column $col");
	    $rows = $tb->rows();	    
	    if ($tb->hasColumn(name => $col)) {
		$co2 = $tb->column(name => $col);
		$v2 = $co2->data();
		$type = $co2->type();
		$ielems = $co2->elements();	    
		for ( $i=0 ; $i < $rows ; $i++ ) {		
		    #
		    # types: 1=bool, 2=Int8, 3=Uint16, 4=Int16,
		    #        5=Uint32, 6=Int32, 7=Real32, 8=Real64, 
		    #        9=DString
		    #
		    # 	if ($type == 7) { # real32
		    #	    $elems = $co2->elements();
		    #	    $new{$name}{$col} = $v2->value($i*$elems);
		    #
		    if ($type == 9) {  # string column
			$new[$i][$c] = getStringCell($co2,$i);
		    } else {
			$msg = "Column ".$col." must be type STRING";
			$code = $col."NotString";
			chomp ($msg,$code);
		      SAS::error($code,$msg);
		    }
		    $co2->release();   # release current column
		}
	    } else {
		$msg = "No ".$col." column in ".$extname." extension";
		$code = "No".$col."Column";
		chomp ($msg,$code);
	      SAS::warning($code,$msg);
		for ( $i=0 ; $i < $rows ; $i++ ) {		
		    $new[$i][$c] = "---";  
		}
	    }
	    $c++;
	}
	$tb->release();
    } else {
	$msg = "No ".$extname." extension in file, no PPS links written";
	$code = "No".$extname;
	chomp ($msg,$code);
      SAS::warning($code,$msg);
    }
    return (\@new);
}
#
# end read_ppsbot
#
#-- 
# WRITE_HTML_LINKS
#--
sub write_HTML_links {   # write links to other summ pages
  
  my ($fh,$inst,@links) = @_;
  my ($key,$val,$tab,$name,%hash);

  my $bordercol="#000099";
  my $highback="#000099";

  @arr = ('OBS Summary','PPS Summary','EPIC Summary','OM Summary','RGS Summary','Catalogue Summary');
  %hash = ('OBX\d\d\dS','OBS Summary','EP','EPIC Summary','OM','OM Summary','RG','RGS Summary','PPSSUM','PPS Summary','CA','Catalogue Summary');
  %newhash = ('OBS Summary','zz','EPIC Summary','zz','OM Summary','zz','RGS Summary','zz','PPS Summary','zz','Catalogue Summary','zz');
  #
  # search thru @links for proper page
  #

 GREPS: foreach $grepval (keys %hash) {  #SUMMAR/EP/OM/RG/PPSSUM/CA
     chomp ($grepval);
   LINKS: foreach $name (@links) {
       chomp ($name);
       if ($name =~ /$grepval/) {
	   $newkey = $hash{$grepval};
	   $newval = $name;
	   $newhash{$newkey} = $newval;
	   next GREPS;
       } 
   } 
 }
  # Form menu bar
  
  &write_generic($fh,"<table width=\"90\%\" align=\"center\" border=\"0\" cellspacing=\"0\" cellpadding=\"3\">\n");
  &write_generic($fh,"\t<tr><td colspan=\"6\">&nbsp;</td></tr>\n");
  &write_generic($fh,"\t<tr align=\"center\">\n");

  foreach $key (@arr) {
      my $label=$key;
      my $line=($key=~/$inst/i)?
	  "<td width=\"15\%\" align=\"center\">$label</td>":
             (!($newhash{$key}=~/zz/))?"<td width=\"15\%\" align=\"center\"><a href=\"$newhash{$key}\">$label</a></td>":
		  "<td width=\"15\%\" align=\"center\"><font color=\"#cccccc\">$label</font></td>";
      &write_generic($fh,"\t\t$line\n");
  }

  &write_generic($fh,"\t</tr>\n");
  &write_generic($fh,"\t<tr><td colspan=\"6\">&nbsp;</td></tr>\n");
  &write_generic($fh,"</table>\n");
}
# 
# end write_html_links
#
#-- 
# WRITE_HTML_INSTSUMM
#--
#sub write_HTML_instsumm {   # write the INSTSUMM info
#  
#  my ($fh,$hash,$isumm,$inst) = @_;
#  my ($key,$val,$tab,$name,@keys);
#  my ($dosumm) = 1;
#  
#  shift (@$isumm); # shift off the INSTRUMENT, since it's the key to $hash
#  
#  if ($inst =~ /E/) {
#    @keys = ('EMOS1','EMOS2','EPN');
#  } elsif ($inst =~ /O/) {
#    @keys = ('OM');
#  } elsif ($inst =~ /R/) {
#    @keys = ('RGS1','RGS2');
#  } else {
#    $dosumm = 0;
#  }
#  
#  my (@cols) = ("Instrument" , "Active" , "Nexp" , "Priority");
#  
#  if ($dosumm) {
#    $name = "Instrument Information:";
#    chomp $name;
#    &write_generic($fh,"<font SIZE=+1>$name</font><br />\n");
#    &write_generic($fh,"<pre>\n");
#    
#    $line = sprintf("%-10s    %6s      %4s    %8s",$cols[0],$cols[1],$cols[2],$cols[3]);
#    chomp $line;
#    &write_generic($fh,"$line\n");
#    &write_generic($fh,"<br />\n");
#    
#    for ($i = 0 ; $i < @keys ; $i++ ) {
#      $inst = $keys[$i];                    # write INSTRUMENT
#      chomp $inst;
#      $act = $hash->{$inst}{@$isumm[0]};    # write ACTIVE flag
#      $act = "Y";  # !!!fix later when ACTIVE isn't a "L" column!
#      $nexp = $hash->{$inst}{@$isumm[1]};   # write NEXPOSURE
#      $prio = $hash->{$inst}{@$isumm[2]};   # write PRIORITY
#      chomp($act,$nexp,$prio);
#      $line = sprintf(" %6s         %2s     %6s       %2s",$inst,$act,$nexp,$prio);
#      &write_generic($fh,"$line\n");
#    }	     
#    &write_generic($fh,"</pre>\n");     # end preformated table
#    &write_generic($fh,"<hr />\n");       # line
#  }
#}
# 
# end write_html_instsumm
#
#-- 
# WRITE_HTML_table
#--
sub write_HTML_table {          # write out a sorted table to HTML
    my ($fh,$dataarray,$colnamearray,$top,$bot,$one,$two,$tre,$tablename,$expid,$inst) = @_;
    my ($key,$val,$tab,$name);
    my ($lineout);
    my (@colnames);
    my $nrows = scalar(@$dataarray);    # nrows is number of files
    my $ncols = scalar(@$bot);
    my $tcols = scalar(@$top);

  SAS::message($SAS::AppMsg,$SAS::SparseMsg,"$tablename will have $nrows rows and $ncols columns");
    
#
# sort the incoming data array on the three parameters (can be 0,0,0)
# WARNING: very friggin' SLOW!
#

    @sortedarray = sort { $a->[$one] cmp $b->[$one] || $a->[$two] cmp $b->[$two] || $a->[$tre] cmp $b->[$tre]} @$dataarray;

#    @sortedarray = sort { $a->[0] cmp $b->[0] || $a->[7] cmp $b->[7]} @$dataarray;
    
    chomp($tablename);

    &write_generic($fh,"<h3>$tablename</h3>\n");

    $prev = "-1";
    $prev_inst = "-1";

    &write_generic($fh,"<!-- S1 -->\n");
    for ($row=0;$row<$nrows;$row++) {
		for($col=0;$col<$ncols;$col++) {     
	   		$column = @$colnamearray[@$bot[$col]];
	    	chomp $column;
	    	$val = $sortedarray[$row][@$bot[$col]];  # current column value
	    	$val = $bandlabel{$val} if ($column eq "BAND");
	    	chomp ($val);
	    	$val = &cleanval($val); # clean whitespace and truncate ".000"
	    	if ($row == 0) {        # only do once
				if ($shorts{$column}) {
		    		$newcol = $shorts{$column};    # shorten the col names
				} 
				else {
		    		$newcol = $column;
				}
				chomp $newcol;	  
				push (@colnames,$newcol);          # new array of col names
				$col_len[$col] = length($newcol);  # width of column 'name'
	    	}
	    	$val = "-" x $col_len[$col] unless length($val);
	    	if (length ($val) > $col_len[$col]) {  # width of column VALUE
				$col_len[$col] = length($val);     # > width of column NAME
	    	}
	    	$all[$row][$col] = $val;
		}
    }
    &write_generic($fh,"<!-- S2 -->\n");
    for ($row=0;$row<$nrows;$row++) {
		for($col=0;$col<$tcols;$col++) {     
	    	$column = @$colnamearray[@$top[$col]];
	    	chomp $column;
	    	$val = $sortedarray[$row][@$top[$col]];
	    	$val = &cleanval($val);
	    	if ($row == 0) {        # only do once
				if ($shorts{$column}) {
		    		$newcol = $short1{$column};    # shorten the col names
				} 
				else {
		    		$newcol = $column;
				}
				chomp $newcol;	  
				push (@topcol,$newcol);          # new array of col names
				$tcol_len[$col] = length($newcol);  # width of column 'name'
	    	}
	    	if (length ($val) > $tcol_len[$col]) {  # width of column VALUE
				$tcol_len[$col] = length($val);     # > width of column NAME
	    	}
		}
    }
#
# fix format for top of page 
#
    &write_generic($fh,"<!-- S3 -->\n");
    &write_generic($fh,"<table width=\"100\%\" cellpadding=\"2\" border=\"0\" cellspacing=\"1\">\n");
    &write_generic($fh,"\t<tr align=\"left\">\n");
    for ($i=0;$i<$tcols;$i++) { 
	
	if ($short1{@$colnamearray[@$top[$i]]}) {
	    $name = $short1{@$colnamearray[@$top[$i]]};
	} 
	else {
	    $name = @$colnamearray[@$top[$i]];
	}

	&write_generic($fh,"\t\t<th>$name</th>\n");
    }
    &write_generic($fh,"\t</tr>\n");
    &write_generic($fh,"<!-- S4 -->\n");
    for ($row=0;$row<$nrows;$row++) {
	    if (($sortedarray[$row][$expid] ne $prev) || ($sortedarray[$row][$inst] ne $prev_inst)) {  # new exposure ID
	        my $line="\t<tr align=\"left\">\n";
	        for ($tcol=0;$tcol<$tcols;$tcol++) {
                # name is column name, namevar is column variable name
	            $name = $short1{@$colnamearray[@$top[$tcol]]};
		        $namevar = $sortedarray[$row][@$top[$tcol]];
		        $exp = $sortedarray[$row][$expid];  # anchor on expid
		        chomp ($name,$namevar,$exp);
		        $namevar = &cleanval($namevar);
		        if  ($name=~/Instrument|Exp/) {   # anchor at Inst & expid
		            $namevar = "<a href=\"$outfile\#$exp\">$namevar</a>";

		        } 
		        $prev = $sortedarray[$row][$expid];
		        $prev_inst = $sortedarray[$row][$inst];
		        $line .= "\t\t<td>$namevar</td>\n";
	        }
	        &write_generic($fh,"$line\t</tr>\n");
	    }
    }

    &write_generic($fh,"</table>\n");


    $prev = "-1";   
    &write_generic($fh,"&nbsp;<br />\n");
    &write_generic($fh,"<hr noshade=\"noshade\" />\n");
    &write_generic($fh,"<!-- E4 -->\n");
#
# fix format statement for exposure records printing below
#
    $lineout = "%-".$col_len[0]."s";
    for ($i=1;$i<$ncols;$i++) {
	$col_len[$i] = $col_len[$i] + 2;
	$lineout .= "  %-"."$col_len[$i]"."s";
    }

#
# now do the individual exposure records...
# 
    if ($in =~ /EPIC|RGS|OM/) {
	&write_generic($fh,"<h3>$in Exposure-Specific Products</h3>\n");
    }

    $prev = "-1";   
    $prev_inst = "-1";   
    for ($row=0;$row<$nrows;$row++) {
	    if (($sortedarray[$row][$expid] ne $prev) || ($sortedarray[$row][$inst] ne $prev_inst)) {  # new exposure ID
		    &write_generic($fh,"&nbsp;<br />\n") if ($row>0);
	    	&write_generic($fh,"<hr noshade=\"noshade\" />\n") if ($row>0);
	    	&write_generic($fh,"<table width=\"100\%\" cellpadding=\"2\" border=\"0\" cellspacing=\"1\">\n");
	    	my ($topline1,$topline2);
	    	$topline1 = $topline2 = "\t<tr>\n";
	    	for ($tcol=0;$tcol<$tcols;$tcol++) {
                # name is column name, namevar is column variable name
				$name = $shorts{@$colnamearray[@$top[$tcol]]};
				$namevar = $sortedarray[$row][@$top[$tcol]];
				$exp = $sortedarray[$row][$expid];  # anchor on expid
				chomp ($name,$namevar,$exp);
				$namevar = &cleanval($namevar);
				$name = &cleanval($name);
				if ($name =~ /Instrument$/) {   # anchor at Instrument
		    		$namevar = "<a name=\"$exp\">$namevar</a>";
#		  SAS::message($SAS::AppMsg,$SAS::SparseMsg,"name is $name");
				} 
				$topline1 .= "\t\t<th align=\"left\">$name</th>\n";
				$topline2 .= "\t\t<td align=\"left\">$namevar</td>\n";
	    	}
	    	&write_generic($fh,"$topline1\t</tr>\n");
	    	&write_generic($fh,"$topline2\t</tr>\n");
	    	&write_generic($fh,"</table>\n");
	    	&write_generic($fh,"&nbsp;<br />\n");

	    	$line = sprintf("$lineout",@colnames);
	    	&write_generic($fh,"<h4><pre>$line</pre></h4>\n");
	    
	  		SAS::message($SAS::AppMsg,$SAS::SparseMsg,"processing new exposure: $exp");
	    	$prev = $sortedarray[$row][$expid];
		    $prev_inst = $sortedarray[$row][$inst];

		}
		for ($col=0;$col<$ncols;$col++) {
	 	   	$val = $all[$row][$col];
	    	if ($col == 0) {  # filename column
				$layout = "%-".$col_len[0]."s";
				$padname = sprintf("$layout",$val);
				if ($val =~ /(HTM)|(PNG)|(PSZ)|(PDF)/) {
		    		$val = href($padname,$padname);  # hotlink
				} 
				else {
		    		$val = $padname;
				}
	    	}
	    	push (@roww,$val);
		}
		$line = sprintf("$lineout",@roww);
#	&write_generic($fh,"<pre>$line</pre>\n");
		&write_generic($fh,"<pre>$line</pre>");
		undef @roww;
    }
    &write_generic($fh,"<br />\n");
    &write_generic($fh,"<hr noshade=\"noshade\" />\n");
    undef @col_len;
}
# 
# end write_html_table
#
#-- 
# WRITE_HTML_multi
#--
sub write_HTML_multi {          # write out a sorted table to HTML
    my ($fh,$dataarray,$colnamearray,$top,$bot,$one,$tablename) = @_;
    my ($key,$val,$tab,$name);
    my ($lineout);
    my (@colnames);
    my $nrows = scalar(@$dataarray);    # nrows is number of files
    my $tcols = scalar(@$top);
    my $bcols = scalar(@$bot);
    
    @sortedarray = sort { $a->[0] cmp $b->[0] } @$dataarray;
    
    chomp($tablename);
    &write_generic($fh,"<h3>$tablename</h3>\n");
    #
    # @$colnamearray is the array of all column names in extension
    #
#
# first write out the top part:
#    
#    foreach $value (@$top) { 
#	$column = @$colnamearray[$value];
#	if ($shorts{$column}) {
#	    $column = $shorts{$column};
#	}
#	chomp $column;
#	$val = $sortedarray[0][$value];
#	chomp ($val);
#	$val = &cleanval($val); # clean whitespace and truncate ".000"	
#	$layout = "%-12s";
#	$padname = sprintf("$layout",$column);
#	$line = $padname.": ".$val;
#      SAS::message($SAS::AppMsg,$SAS::SparseMsg,"line is $line");
#	&write_generic($fh,"$line<br />\n");
#    }

    my ($col1_head,$col2_head);
    if ($shorts{@$colnamearray[0]}) {
	$col1_head = $shorts{@$colnamearray[0]};    # shorten the col names
    } else {
	$col1_head = @$colnamearray[0]
    }
   if ($shorts{@$colnamearray[1]}) {
	$col2_head = $shorts{@$colnamearray[1]};    # shorten the col names
    } else {
	$col2_head = @$colnamearray[1]
    }

    $layout = "%-33s %-40s";

    $line = sprintf("$layout",$col1_head,$col2_head);
 
    &write_generic($fh,"<h4><pre>$line</pre></h4>\n");
    $line = "";
    for ($row=0;$row<$nrows;$row++) {

	$name = $sortedarray[$row][0];
	$content = $sortedarray[$row][1];
	$padout = "%-33s";    # pad out filenames for now
	$padname = sprintf("$padout",$name);
	if ($content =~ /SUMM|OBSER/) {
	    if ($padname =~ /(HTM)|(PNG)|(PSZ)|(PDF)/) {
		$padname = href($padname,$padname);  # hotlink
	    } else {
		$val = $padname;
	    }
	    $line = sprintf("$layout",$padname,$content);
#Bookmark
	    &write_generic($fh,"<pre>$line</pre>\n") if ($line !~ /\d+SUMMAR/);
	}
    }
    &write_generic($fh,"<hr noshade=\"noshade\" />\n");
}
# 
# end write_html_multi
#
#-- 
# WRITE_HTML_pps
#--
sub write_HTML_pps {          # write out a sorted table to HTML
    
#    &write_HTML_pps($htmlfile,$pps_ref,\@PPSOBS,$pps_bot,\@summnames,\@ppsexp);
    my ($fh,$dataarray,$set_key,$colnamearray,$bot,$summnames,$cols) = @_;
    my ($key,$val,$tab,$name);
    my ($lineout);
    my (@colnames);
    my $nrows = scalar(@$dataarray);    # nrows is number of files
    my $tcols = scalar(@$colnamearray);
    $prevEXP = -1;
    $prevINST = -1;
#    
    @sortedbot = sort { $a->[0] cmp $b->[0] } @$bot;
    my $brows = scalar(@sortedbot);
#
# sort on instrument, then start time (expid should fall into place, but
# are often not in ascending order, so use START TIME)
#
    @sortedarray = sort { $a->[0] cmp $b->[0] || $a->[3] cmp $b->[3]} @$dataarray;    

  SAS::message($SAS::AppMsg,$SAS::SparseMsg,"Writing PPS Summary Page Records.");

    #
    # @$colnamearray is the array of all column names in extension
    #

    &write_generic($fh,"<h3>Exposures processed by PPS</h3>\n");

    @lay = (6,6,15,20,10,20);
    $layout = "%-6s %-6s %-20s %-20s %-10s %-20s";
    $line = sprintf("$layout","Inst","ExpID","Mode","Start time","Duration","Stop time");
    &write_generic($fh,"<h4><pre>$line</pre></h4>\n");

    %exts = ("EMOS1" => "$summs{EPICOBS}",
	     "EMOS2" => "$summs{EPICOBS}",
	     "EPN" => "$summs{EPICOBS}",
	     "RGS1" => "$summs{RGSOBS}",
	     "RGS2" => "$summs{RGSOBS}",
	     "OM" => "$summs{OMOBS}",
	     "CAT" => "$summs{CATOBS}"
	     );
    for ($row=0;$row<$nrows;$row++) {
#
# if current exposure ID != previous, or new instrument
# write a new entry
#
	$inst = $sortedarray[$row][0];
	$expid = $sortedarray[$row][1];
	if (($expid ne $prevEXP) or 
	    ($inst  ne $prevINST)) {
	    for ($col=0;$col<6;$col++) {
		$val = $sortedarray[$row][$col];
#		chomp ($val);
		$val = &cleanval($val); # clean whitespace and truncate ".000"
 		$lyt = "%-".$lay[$col]."s"; 		
		$val = sprintf("$lyt",$val);		
		if ($col < 2) {
#
# add hotlink if instrument or expid
#
		    $val = "<a href=\"$exts{$inst}\#$expid\">$val</a>";
		}
#	      SAS::message($SAS::AppMsg,$SAS::SparseMsg,"val is $val");
		push (@all,$val);
	    }
	    $line = sprintf("$layout",@all);
	    &write_generic($fh,"<pre>$line</pre>\n");
	    undef @all;
	}
	$prevEXP = $sortedarray[$row][1];
	$prevINST = $sortedarray[$row][0];

    }

    &write_generic($fh,"<br />\n");
    &write_generic($fh,"<hr noshade=\"noshade\">\n");
    &write_generic($fh,"<h3>Generic PPS Products</h3>\n");

    $layout = "%-33s %-40s";
    $line = sprintf("$layout","Filename","Content");
    &write_generic($fh,"<h4><pre>$line</pre></h4>\n") unless ($brows<2);

    for ($row=0;$row<$brows;$row++) {
	$name = $sortedbot[$row][0];
	$content = $sortedbot[$row][1];
	$padout = "%-33s";    # pad out filenames for now
	$padname = sprintf("$padout",$name);
	if ($content =~ /SUMM|OBSER/) {
	    if ($padname =~ /(HTM)|(PNG)|(PSZ)|(PDF)/) {
		$padname = href($padname,$padname);  # hotlink
	    }
	}
	$line = sprintf("$layout",$padname,$content);
	&write_generic($fh,"<pre>$line</pre>\n") unless (($name =~ /SUMMAR/) || ($name =~ /\.PNG$/));
    }
    &write_generic($fh,"<br />\n");
    &write_generic($fh,"<hr noshade=\"noshade\">\n");
}
# 
# end write_html_multi
#
sub cleanval {

    my ($value) = @_;
    chomp $value;

#  SAS::message($SAS::AppMsg,$SAS::SparseMsg,"val is $value");

    $value =~ s/^\s+//;      # leading white  (more shortening)
    $value =~ s/\s+$//;      # trailing white
    $value =~ s/\.000+$//;   # trailing ".000" in time cols
    $value =~ s/\s+/ /g;     # no middle white +
    return ($value);
}
#--
#read_attributes
#--
sub read_attributes{
  SAS::message($SAS::AppMsg,$SAS::SparseMsg,"Reading Attributes");
    my ($ds,$attr) = @_;
    my ($name,$p);

    foreach $name (keys(%$attr)) {
	if ($ds->hasAttribute(name => "$name")) {
	    $p = $ds->attribute(name => "$name");
	    $attr->{"$name"}{"value"} = &cleanval($p->asString());
	}
	else {
	    $attr->{"$name"}{"value"} = " "; 
	}
    }
#    return ($attr);
}
#
# end read_attributes
#
sub generate_sub_title {
    my ($attr,$order) = @_;
    my $table_start = "<table border=\"0\" width=\"100\%\">\n";
    my $table_end = "</table>\n";
    my $row_start =  "\t<tr align=\"left\">\n";
    my $row_end = "\t</tr>\n";
    my ($st1) = "$row_start";
    my ($st2) = "$row_start";

    foreach $key (@$order) {
	$st1 .= "\t\t<th>$attr->{$key}{text}</th>\n";
	$attr->{$key}{value} = "*" if (!($attr->{$key}{value}));
	$st2 .= "\t\t<td>$attr->{$key}{value}</td>\n";
    }
    $st1 .= $row_end;
    $st2 .= $row_end;
    return  "$table_start $st1  $st2 $table_end";

}

sub get_key {
    my ($inp)=@_;
    my @key = split /\.+$/ , $inp; 
    return $key[0];
}
